import copy
import numpy as np
import pandas as pd
import cvxpy as cp
from scipy import stats
import sklearn.linear_model as lm
from . import utilities
from .utilities import parse_dist, BatchedCategorical, vrange
# typing
import sklearn.base
from typing import Optional, Union

TOL = 1e-6
DEFAULT_ALPHAS = np.log(np.logspace(0.0001, 100, base=np.e, num=10))
DEFAULT_ALPHAS = np.concatenate([DEFAULT_ALPHAS, np.array([1000, 10000])])


def parse_model_type(model_type, discrete):
	# return non-strings
	if not isinstance(model_type, str):
		return model_type
	# parse
	model_type = model_type.lower()
	if not discrete:
		if model_type == 'ridge':
			return lm.RidgeCV
		elif model_type == 'ols':
			return lm.LinearRegression
		elif model_type == 'lasso':
			return lm.LassoCV
		elif model_type in ['elastic', 'elasticnet']:
			return lm.ElasticNetCV
		elif model_type in ['rf', 'randomforest']:
			import sklearn.ensemble
			return sklearn.ensemble.RandomForestRegressor
		elif model_type == 'knn':
			import sklearn.neighbors
			return sklearn.neighbors.KNeighborsRegressor
		else:
			raise ValueError(f"Unrecognized model_type={model_type}")
	else:
		if model_type in ['logistic', 'ridge', 'lasso', 'elastic', 'elasticnet']:
			return lm.LogisticRegressionCV
		elif model_type in ['monotone_logistic', 'logistic_monotone']:
			return MonotoneLogisticReg
		elif model_type in ['rf', 'randomforest']:
			import sklearn.ensemble
			return sklearn.ensemble.RandomForestClassifier
		elif model_type == 'knn':
			import sklearn.neighbors
			return sklearn.neighbors.KNeighborsClassifier
		else:
			raise ValueError(f"Unrecognized model_type={model_type}")

def create_folds(n, nfolds):
	splits = np.linspace(0, n, nfolds+1).astype(int)
	starts = splits[0:-1]
	ends = splits[1:]
	return starts, ends

def _enforce_monotonicity(
	y0_dists: Union[BatchedCategorical, np.array], 
	y1_dists: Union[BatchedCategorical, np.array],
	margin=0.005,
):
	"""
	For a binary variable, enforces the monotonicity constraint
	that P(Y_i(0) = 1) <= P(Y_i(1) = 1).

	Parameters
	----------
	y0_dists : BatchedCategorical | np.array
		Either an array whose ith element is P(Y_i(0) = 1)
		or a BatchedCategorical distribution.
	y1_dists : BatchedCategorical | np.array
		Either an array whose ith element is P(Y_i(1) = 1)
		or a BatchedCategorical distribution.
	margin : float
		Minimum distance between P(Y_i(0) = 1) and P(Y_i(1) = 1).

	Returns
	-------
	y0_dists : BatchedCategorical | np.array
		Either an array whose ith element is P(Y_i(0) = 1)
		or a BatchedCategorical distribution.
	y1_dists : BatchedCategorical | np.array
		Either an array whose ith element is P(Y_i(1) = 1)
		or a BatchedCategorical distribution.

	Notes
	-----
	The output format is the same as the input format.
	"""
	# Convert to np array
	if isinstance(y0_dists, BatchedCategorical):
		p0s = y0_dists.probs[:, 1].copy()
	else:
		p0s = y0_dists.copy()

	if isinstance(y1_dists, BatchedCategorical):
		p1s = y1_dists.probs[:, 1].copy()
	else:
		p1s = y1_dists.copy()

	# Clip
	p0s = np.minimum(1-TOL, np.maximum(TOL, p0s))
	p1s = np.minimum(1-TOL, np.maximum(TOL, p1s))
	# enforce monotonicity
	flags = p0s > p1s - margin
	avg = (p0s[flags] + p1s[flags]) / 2
	p0s[flags] = np.maximum(TOL, avg-margin)
	p1s[flags] = np.minimum(1-TOL, avg+margin)
	# ensure correct format
	if isinstance(y0_dists, BatchedCategorical):
		p0s = BatchedCategorical.from_binary_probs(p0s)
	if isinstance(y1_dists, BatchedCategorical):
		p1s = BatchedCategorical.from_binary_probs(p1s)
	# Return
	return p0s, p1s

def _evaluate_model_predictions(y, haty, tol=1e-9):
	# Preliminaries
	n = len(y)
	resids = y - haty
	ybar_loo = (n * y.mean() - y) / (n-1)
	resids_null = y - ybar_loo
	# RMSE calculations
	rmse = np.sqrt(np.mean(resids**2))
	rmse_null = np.sqrt(np.mean(resids_null**2))
	oos_r2 = 1 - rmse**2 / rmse_null**2
	# Accuracy/likelihood/MAE calculations
	if set(list(np.unique(y))) == set([0,1]):
		y = y.astype(int)
		# accuracy
		haty_round = np.round(np.clip(haty, 0, 1)).astype(int)
		accuracy = (haty_round == y).astype(float)
		accuracy[haty == 0.5] = 0.5
		accuracy = np.mean(accuracy)
		# null model accuracy
		null_accuracy = (np.round(ybar_loo).astype(int) == y).astype(float)
		null_accuracy[ybar_loo == 0.5] = 0.5
		null_accuracy = np.mean(null_accuracy)
		# likelihood
		haty = np.clip(haty, tol, 1-tol)
		lls = np.log(haty) * y + np.log(1-haty) * (1 - y)
		ybar_loo = np.clip(ybar_loo, tol, 1-tol)
		null_lls = np.log(ybar_loo) * y + np.log(1-ybar_loo) * (1-y)
		# take geometric mean
		likelihood_geom = np.exp(lls.mean())
		null_likelihood_geom = np.exp(null_lls.mean())
		return pd.DataFrame(
			np.array([
				[oos_r2, 0],
				[accuracy, null_accuracy],
				[likelihood_geom, null_likelihood_geom]
			]),
			index=['Out-of-sample R^2', 'Accuracy', 'Likelihood (geom. mean)'],
			columns=['Model', 'No covariates'],
		)
	else:   
		mae = np.mean(np.abs(resids))
		mae_null = np.mean(np.abs(resids_null))
		return pd.DataFrame(
			np.array([
				[oos_r2, 0],
				[rmse, rmse_null],
				[mae, mae_null]
			]),
			index=['Out-of-sample R^2', 'RMSE', 'MAE'],
			columns=['Model', 'No covariates'],
		)

def ridge_loo_resids(
	features, y, ridge_cv_model
):
	# Ensure we add an intercept which is unregularized
	n, p = features.shape
	if ridge_cv_model.fit_intercept:
		# This is not *exact* if we fit an intercept
		# because it assumes the intercept doesn't change
		# but it's realistically pretty close
		X_offset = features.mean(axis=0)
		y_offset = y.mean(axis=0)
		y = y - y_offset; features = features - X_offset
	# Regularization strength
	try:
		alpha = ridge_cv_model.alpha_
	except:
		alpha = ridge_cv_model.alpha
	# Predictions
	Q = features.T @ features + alpha * np.eye(p)
	Qinv = np.linalg.inv(Q)
	hatbeta = Qinv @ features.T @ y
	preds = features @ hatbeta
	# Leave one out residuals
	scales = np.sum((features @ Qinv) * features, axis=1)
	loo_resids = (y - preds) / (1 - scales)
	return loo_resids

class DistReg:
	"""
	A generic class for distributional regressions, meant for subclassing.

	Parameters
	----------
	how_transform : str
		Str specifying how to transform the features before fitting
		the underlying model. One of several options:

		- 'identity': does not transform the features
		- 'intercept': adds an intercept
		- 'interactions' : adds treatment-covariate interactions

		The default is ``interactions``.

	Notes
	-----
	To inherit from this class, simply define the ``fit``
	and ``predict`` functions, ensuring that they match the 
	signature in the docs for this class.
	"""

	def __init__(self, how_transform):
		self.how_transform = how_transform


	def fit(
		self,
		W: np.array,
		X: np.array,
		y: np.array,
		Z: Optional[np.array]=None,
		sample_weight: Optional[np.array]=None,
	):
		"""
		Fits model on the data.

		Parameters
		----------
		W : np.array
			n-length array of binary treatment indicators.
		X : np.array
			(n, p)-shaped array of covariates.
		y : np.array
			n-length array of outcome measurements.
		Z : np.array
			Optional n-length array of instrument values
			for the instrumental variables setting.
		sample_weight : np.array
			Optional n-length array of weights to use when
			fitting the underlying model.

		Notes
		-----
		Not all model_types can be trained using sample_weight.
		For some models (e.g. KNNRegressors), including
		sample_weight will yield an error.
		"""
		raise NotImplementedError()

	def feature_transform(self, W: np.array, X: np.array, Z: Optional[np.array]=None):
		"""
		Transforms the features before feeding them to the base model.

		Parameters
		----------
		X : np.array
			(n, p)-shaped array of covariates.
		W : np.array
			n-length array of treatment/exposure indicators.
		Z : np.array
			Optional n-length array of binary instruments
			for the instrumental variables setting.

		Returns
		-------
		features : np.array
			(n, d)-shaped array of features.
		"""
		# Transformations based on W
		if self.how_transform in ['none', 'identity']:
			features = np.concatenate([W.reshape(-1, 1), X],axis=1)
		elif self.how_transform in ['intercept']:
			features = np.concatenate([W.reshape(-1, 1), np.ones((len(X), 1)), X],axis=1)
		elif self.how_transform in [
			'int', 'interaction', 'interactions'
		]:
			features = np.concatenate([
				W.reshape(-1, 1), W.reshape(-1, 1) * X, X], axis=1
			)
		else:
			raise ValueError(f"Unrecognized transformation {self.how_transform}")

		# Add Z. Note that we cannot add interactions with W
		if Z is None:
			return features
		else:
			Z = Z.reshape(-1, 1)
			if self.how_transform in ['int', 'interaction', 'interactions']:
				return np.concatenate([features, Z, Z * X], axis=1)
			else:
				return np.concatenate([features, Z], axis=1)

	def features_to_WX(self, features: np.array):
		"""
		Inverse of feature_transform.
		"""
		# Raise error
		if self._iv:
			raise NotImplementedError("features_to_XW not implemented for IV regressions.")
		# Reverse transform
		if self.how_transform in ['none', 'identity']:
			return features[:, 0], features[:, 1:]
		elif self.how_transform in ['intercept']:
			return features[:, 0], features[:, 2:]
		else:
			p = int((features.shape[1] - 1) / 2)
			return features[:, 0], features[:, 1:(p+1)]

	def predict(
		self, 
		X: np.array,
		W: np.array,
		Z: np.array,
	):
		"""
		Predicts the conditional law of the outcome.

		Parameters
		----------
		X : np.array
			(n, p)-shaped array of covariates.
		W : np.array
			Optional n-length array of binary treatment indicators.
		Z : np.array
			Optional n-length array of binary instruments
			for the instrumental variables setting.

		Returns
		-------
		y_dists : stats.rv_continuous / stats.rv_discrete
			batched scipy distribution of shape (n,) where the ith
			distribution is the conditional law of :math:`Y_i | X_i, W_i, Z_i`
			(without conditioning on :math:`Z_i` if ``Z`` is not provided).
		"""
		raise NotImplementedError()

	def predict_counterfactuals(self, X: np.array):
		"""
		Predicts counterfactual distributions of Y (outcome).

		Parameters
		----------
		X : np.array
			(n,p)-shaped design matrix.

		Returns
		-------
		y0_dists : np.array
			y0_dists[i] = the law of :math:`Y_i(0) | X_i`.
			Only returned in a setting without instrumental variables.
		y1_dists : np.array
			y1_dists[i] = the law of :math:`Y_i(1) | X_i`.
			Only returned in a setting without instrumental variables.
		ydists : list
			Only returned if trained using instrumental variables.
			Then ydists[z][w] is a batched scipy distribution
			array whose ith element represents the law of
			:math:`Y_i(w) | X_i, W_i(z) = w`.
		"""
		n = len(X)
		# Case 1: instrument variables (4 counterfactuals)
		if self._iv:
			ydists = [[], []]
			for z in [0,1]:
				y0_dists = self.predict(X=X, W=np.zeros(n), Z=z*np.ones(n))
				y1_dists = self.predict(X=X, W=np.ones(n), Z=z*np.ones(n))
				ydists[z] = [y0_dists, y1_dists]
			return ydists
		# Case 2: intent to treat (2 counterfactuals)
		else:
			y0_dists = self.predict(X=X, W=np.zeros(n))
			y1_dists = self.predict(X=X, W=np.ones(n))
			return y0_dists, y1_dists

class ModelSelector():
	"""
	Class which selects a distributional regression model.

	Notes
	-----
	This class selects the model with the best out-of-sample
	R^2. However, by wrapping this class, other selection metrics
	can be used (simply overwrite the ``select_model`` method).
	"""
	def __init__(self):
		pass

	def select_model(
		self,
		models: list[DistReg],
		W: np.array,
		X: np.array,
		y: np.array,
		Z: Optional[np.array]=None,
		sample_weight: Optional[np.array]=None,
		propensities: Optional[np.array]=None,
		**kwargs
	) -> DistReg:
		"""
		Parameters
		----------
		models : list
			list of ``dist_reg.DistReg`` classes.
		W : np.array
			n-length array of binary treatment indicators.
		X : np.array
			(n, p)-shaped array of covariates.
		y : np.array
			n-length array of outcome measurements.
		Z : np.array
			Optional n-length array of binary instrument values
			for the instrumental variables setting.
		propensities : np.array
			Optional n-length array of propensity scores. 
			This argument is only used when ``model_selector``
			is provided.
		sample_weight : np.array
			Optional n-length array of weights to use when
			fitting the underlying model.

		Returns
		-------
		selected_model : ``dist_reg.DistReg``
			The selected model.
		"""
		# Evaluate out-of-sample R^2s
		r2s = np.zeros(len(models))
		for i, model in enumerate(models):
			_, _, oos_preds = cross_fit_predictions(
				W=W, X=X, y=y, Z=Z, 
				sample_weight=sample_weight,
				propensities=propensities,
				model=model,
				model_selector=None,
				verbose=False,
			)
			results = _evaluate_model_predictions(
				y, haty=np.concatenate([x.mean() for x in oos_preds])
			)
			r2s[i] = results.loc['Out-of-sample R^2', 'Model'].item()
		# Pick the best-performing model
		return models[np.argmax(r2s)]


def cross_fit_predictions(
	W: np.array,
	X: np.array,
	y: np.array,
	Z: Optional[np.array]=None,
	S: Optional[np.array]=None,
	propensities: Optional[np.array]=None,
	sample_weight: Optional[np.array]=None,
	nfolds: int=5,
	train_on_selections: bool=True,
	model: Optional[Union[list, DistReg]]=None,
	probs_only: bool=False,
	verbose: bool=False,
	model_selector: Optional[ModelSelector]=None,
):
	"""
	Performs cross-fitting for a model inheriting from ``dist_reg.DistReg.``

	Parameters
	----------
	W : np.array
		n-length array of binary treatment indicators.
	X : np.array
		(n, p)-shaped array of covariates.
	y : np.array
		n-length array of outcome measurements.
	Z : np.array
		Optional n-length array of binary instrument values
		for the instrumental variables setting.
	S : np.array
		Optional n-length array of selection indicators.
	propensities : np.array
		Optional n-length array of propensity scores. 
		This argument is only used when ``model_selector``
		is provided.
	sample_weight : np.array
		Optional n-length array of weights to use when
		fitting the underlying model.
	nfolds : int
		Number of cross-fitting folds to use.
	model : DistReg
		instantiation of ``dist_reg.DistReg`` class. This will
		be copied. E.g., 
		``model=dist_reg.CtsDistReg(model_type='ridge', eps_dist="empirical").``
		Alternatively, one may provide a list of
		``dist_reg.DistReg`` classes and a ``model_selector`` which
		adaptively chooses between them.
	train_on_selections : bool
		If True, trains model only on data where S[i] == 1.
	probs_only : bool
		For binary data, returns P(Y = 1 | X, W) instead of a distribution.
		Defaults to False.
	verbose : bool
		If True, provides progress reports.

	Returns
	-------
	y0_dists : list
		list of batched scipy distributions whose shapes sum to n.
		the ith distribution is the out-of-sample estimate of
		the conditional law of Yi(0) | X[i]
	y1_dists : list
		list of batched scipy distributions whose shapes sum to n.
		the ith distribution is the out-of-sample estimate of
		the conditional law of Yi(1) | X[i]
	"""
	# concatenate S to features
	n = len(X)
	if S is None:
		S_flag = False
		S = np.ones(n)
	else:
		S_flag = True
		X = np.concatenate([S.reshape(n, 1), X], axis=1)

	# create folds
	starts, ends = create_folds(n=n, nfolds=nfolds)
	# loop through folds
	counterfactual_preds = []# results for different values of W/Z
	oos_preds = [] # results for true W/Z
	fit_models = []
	for ii in vrange(len(starts), verbose=verbose):
		start = starts[ii]; end = ends[ii]
		# Pick out data from the other folds
		not_in_fold = [i for i in np.arange(n) if i < start or i >= end]
		if train_on_selections:
			opt2 = [i for i in not_in_fold if S[i] == 1]
			if len(opt2) == 0:
				warnings.warn(f"S=0 for all folds except {start}-{end}.")
			else:
				not_in_fold = opt2

		# Possibly perform model selection
		fold_args = dict(
			W=W[not_in_fold], 
			X=X[not_in_fold], 
			y=y[not_in_fold],
			Z=None if Z is None else Z[not_in_fold],
			sample_weight=None if sample_weight is None else sample_weight[not_in_fold],
		)
		if isinstance(model, list):
			if model_selector is None:
				model_selector = ModelSelector()
			reg_model = copy.deepcopy(model_selector.select_model(
				models=model,
				propensities=None if propensities is None else propensities[not_in_fold],
				**fold_args,
			))
		else:
			reg_model = copy.deepcopy(model)
		
		# Fit model
		reg_model.fit(**fold_args)

		# predict and append on this fold
		subX = X[start:end].copy(); 
		if S_flag:
			subX[:, 0] = 1 # set selection = 1 for predictions

		# Actual predictions
		oos_pred = reg_model.predict(
			subX, W[start:end], Z=None if Z is None else Z[start:end]
		)
		counterfactual_pred = reg_model.predict_counterfactuals(subX)

		# Save predictions and model fits
		counterfactual_preds.append(counterfactual_pred)
		oos_preds.append(oos_pred)
		fit_models.append(reg_model)

	# Appropriately concatenate counterfactual predictions. 
	# Requires different code for the IV/non-IV case.
	if Z is not None:
		cf_output = [[np.nan, np.nan], [np.nan, np.nan]]
		for z in [0,1]:
			for w in [0,1]:
				cf_output[z][w] = [x[z][w] for x in counterfactual_preds]
				# Concatenate probabilities if we're only interested in them
				if probs_only:
					cf_output[z][w] = np.concatenate(
						[x.probs[:, 1] for x in cf_output[z][w]]
					)

	else:
		cf_output = [[], []]
		for z in [0,1]:
			cf_output[z] = [x[z] for x in counterfactual_preds]
			# Concatenate probabilities if we're only interested in probabilities
			if probs_only:
				cf_output[z] = np.concatenate([x.probs[:, 1] for x in cf_output[z]])



	return cf_output, fit_models, oos_preds


class CtsDistReg(DistReg):
	"""
	Distributional regression for continuous outcomes.

	Parameters
	----------
	model_type : str or sklearn class
		Str specifying a sklearn model class to use; options include
		'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'. One can
		also directly pass an sklearn class, e.g., 
		``model_type=sklearn.ensemble.KNeighborsRegressor``. 
	how_transform : str
		Str specifying how to transform the features before fitting
		the underlying model. One of several options:

		- 'identity': does not transform the features
		- 'intercept': adds an intercept
		- 'interactions' : adds treatment-covariate interactions

		The default is ``interactions``.
	eps_dist : str
		Str specifying the distribution of the residuals. Options include
		['empirical', gaussian', 'laplace', 'expon', 'tdist', 'skewnorm']. 
		Defaults to ``empirical``, which uses the empirical law of the
		residuals of the training data.
	eps_kwargs : dict
		kwargs to ``utilities.parse_dist`` for the residual scipy distribution
	heterosked_model : str or sklearn class
		Str specifying a sklearn model class to use to estimate 
		Var(Y | X) as a function of X. Options are the same as 
		``model_type.`` Defaults to heterosked_model=None, in which
		case homoskedasticity is assumed (although the final bounds
		will still be valid in the presence of heteroskedasticity).
	heterosked_kwargs : dict
		kwargs for the heterosked model. E.g., 
		if ``heterosked_model=knn``,
		heterosked_kwargs could include ``n_neighbors``.
	**model_kwargs : dict
		kwargs for sklearn base model. E.g., if ``model_type=knn``,
		model_kwargs could include ``n_neighbors``.

	Examples
	--------
	Here we instantiate a model which assumes Gaussianity,
	uses a ridge to make predictions and a lasso to estimate
	the heteroskedasticity pattern: ::

		import numpy as np
		import dualbounds
		import sklearn.linear_model

		# Instantiate dist_reg
		cdreg = dualbounds.dist_reg.CtsDistReg(
			# Arguments for main model
			model_type=sklearn.linear_model.RidgeCV,
			fit_intercept=True,
			gcv_mode='auto',
			# How to estimate the law of the residuals
			eps_dist='gaussian',
			# How to estimate Var(Y | X)
			heterosked_model=sklearn.linear_model.LassoCV,
			heterosked_kwargs=dict(cv=5),
		)

		# Fit
		n, p = 300, 20
		W = np.random.binomial(1, 0.5, n)
		X = np.random.randn(n, p)
		y = np.random.randn(n)
		cdreg.fit(W=W, X=X, y=y)

		# Predict on new X
		m = 10
		Xnew = np.random.randn(m, p)
		y0_preds = cdreg.predict(X=Xnew, W=np.zeros(m))
	"""
	def __init__(
		self,
		model_type: Union[str, sklearn.base.BaseEstimator] = 'ridge',
		how_transform: str = 'interactions',
		eps_dist: str = 'empirical',
		eps_kwargs: Optional[dict]=None,
		heterosked_model: str='none',
		heterosked_kwargs: Optional[dict]=None,
		**model_kwargs,
	):
		# Distribution of residuals
		self.eps_dist = eps_dist
		self.eps_kwargs = {} if eps_kwargs is None else eps_kwargs
		self.use_loo = self.eps_kwargs.pop("use_loo", False)
		# Main predictive model
		self.model_type = parse_model_type(model_type, discrete=False)
		self.model_kwargs = model_kwargs
		self.how_transform = str(how_transform).lower()
		## Heteroskedasticity
		# Parse if we want to fit a model for heteroskedasticity
		if heterosked_model is not None:
			if isinstance(heterosked_model, str):
				heterosked_model = heterosked_model.lower()
			self.heterosked = heterosked_model != 'none'
		else:
			self.heterosked = False
		# If so, parse the model type
		if self.heterosked:
			self.sigma2_model_type = parse_model_type(heterosked_model, discrete=False)
			self.sigma2_model_kwargs = heterosked_kwargs
			if self.sigma2_model_kwargs is None:
				self.sigma2_model_kwargs = {}
		# Default kwargs
		if model_type == 'ridge':
			self.model_kwargs['alphas'] = self.model_kwargs.get("alphas", DEFAULT_ALPHAS)
		

	def fit(
		self, 
		W: np.array,
		X: np.array,
		y: np.array,
		Z: Optional[np.array]=None,
		sample_weight: Optional[np.array]=None
	) -> None:
		self._iv = Z is not None
		self.Wtrain = W
		# fit model
		features = self.feature_transform(W=W, X=X, Z=Z)
		self.model = self.model_type(**self.model_kwargs)
		# Possibly use sample_weight
		fit_kwargs = dict() if sample_weight is None else dict(sample_weight=sample_weight)
		self.model.fit(features, y, **fit_kwargs)

		# compute residuals. Use cheap LOO resids for ridge
		if isinstance(self.model, lm.RidgeCV) and self.use_loo:
			self.resids = ridge_loo_resids(features, y=y, ridge_cv_model=self.model)
		else:
			self.resids = y - self.model.predict(features)

		# Estimate E[Var(Y | X)] using residuals.
		self.hatsigma = np.sqrt(np.mean(self.resids**2))
		if self.heterosked:
			self.sigma2_model = self.sigma2_model_type(**self.sigma2_model_kwargs)
			self.sigma2_model.fit(features, self.resids**2, **fit_kwargs)
			self.hatsigma_preds = np.sqrt(np.clip(
				self.sigma2_model.predict(features),
				0.01 * self.hatsigma**2, 
				np.inf
			))
		else:
			self.hatsigma_preds = self.hatsigma

		# If eps_dist == empirical, fit empirical law
		if self.eps_dist == 'empirical':
			norm_resids = self.resids / self.hatsigma_preds
			self.rvals0 = np.sort(norm_resids[W == 0])
			# Center to ensure we don't scale a small mean by a large amount
			# when heterosked=True.
			if sample_weight is None:
				self.rprobs0 = np.ones(len(self.rvals0)) / len(self.rvals0)
			else:
				self.rprobs0 = sample_weight[W == 0].copy()
				self.rprobs0 /= self.rprobs0.sum()
			self.rvals0 -= self.rvals0 @ self.rprobs0
			# Repeat for treatment
			self.rvals1 = np.sort(norm_resids[W == 1])
			if sample_weight is None:
				self.rprobs1 = np.ones(len(self.rvals1)) / len(self.rvals1)
			else:
				self.rprobs1 = sample_weight[W == 1].copy()
				self.rprobs1 /= self.rprobs1.sum()
			self.rvals1 -= self.rvals1 @ self.rprobs1
			# adjust support size for computational reasons
			adj_kwargs = dict(
				new_nvals=self.eps_kwargs.get("nvals", 90), ymin=y.min(), ymax=y.max()
			)
			self.rvals0, self.rprobs0 = utilities._adjust_support_size_unbatched(
				self.rvals0, self.rprobs0, **adj_kwargs
			)
			self.rvals1, self.rprobs1 = utilities._adjust_support_size_unbatched(
				self.rvals1, self.rprobs1, **adj_kwargs
			)

	def predict(self, X: np.array, W: np.array, Z: Optional[np.array]=None):

		features = self.feature_transform(W, X=X, Z=Z)
		mu = self.model.predict(features)
		# heteroskedasticity
		if self.heterosked:
			scale = np.sqrt(np.clip(
				self.sigma2_model.predict(features),
				0.01 * self.hatsigma**2,
				np.inf,
			))
		else:
			scale = self.hatsigma
		# Option 1: nonparametric law of eps
		if self.eps_dist == 'empirical':
			# law of normalized residuals
			# only use empirical residuals from the same treatment/control group
			vals = np.stack(
				[self.rvals0 if Wi == 0 else self.rvals1 for Wi in W],
				axis=0
			)
			vals = vals * np.array([scale]).reshape(-1, 1) + mu.reshape(-1, 1)
			# Create probs and return
			probs = np.stack(
				[self.rprobs0 if Wi == 0 else self.rprobs1 for Wi in W], 
				axis=0
			)
			return utilities.BatchedCategorical(
				vals=vals, probs=probs
			)
		# Option 2: parametric law of eps, return scipy dists
		else:
			return parse_dist(
				self.eps_dist, mu=mu, sd=scale, **self.eps_kwargs,
			)

class QuantileDistReg(DistReg):
	"""
	A continuous distributional regression based on quantile regression.

	Parameters
	----------
	nquantiles : int
		The number of quantiles to fit quantile regressions for.
		Quantiles are evenly spaced between 0 and 1.
	alphas : list or np.array
		List of l1 regularization strengths to use in the quantile
		regression; the final strength is determined by cross-validation.  
		Default is alphas=[0] (no regularization).
	how_transform : str
		Str specifying how to transform the features before fitting
		the underlying model. One of several options:

		- 'identity': does not transform the features
		- 'intercept': adds an intercept
		- 'interactions' : adds treatment-covariate interactions

		The default is ``interactions``.

	Notes
	-----
	This method is computationally expensive for large datasets.
	"""
	def __init__(
		self, 
		nquantiles: int=50,
		alphas: list=[0],
		how_transform: str='interactions',
	):
		self.nq = nquantiles
		self.how_transform = how_transform
		self.quantiles = np.around(np.linspace(0, 1, self.nq), 8)
		self.probs = self.quantiles[1:] - self.quantiles[0:-1]
		self.alphas = alphas

	def fit(
		self,
		W: np.array, 
		X: np.array, 
		y: np.array, 
		Z: Optional[np.array]=None, 
		sample_weight: Optional[np.array]=None
	) -> None:
		self._iv = Z is not None
		# Pick regularization by using CV lasso reg. strength
		features = self.feature_transform(W=W, X=X, Z=Z)
		if len(self.alphas) > 0:
			from sklearn.metrics import d2_pinball_score, make_scorer
			from sklearn.model_selection import cross_validate
		# Fit many quantiles
		self.model = {}
		self.scores = {}
		self.ymin = y.min()
		self.ymax = y.max()
		fit_kwargs = dict() if sample_weight is None else dict(sample_weight=sample_weight)
		for quantile in self.quantiles:
			if quantile not in [0,1]:
				if len(self.alphas) == 1:
					qr = lm.QuantileRegressor(
						alpha=self.alphas[0], quantile=quantile, 
					)
					qr.fit(features, y, **fit_kwargs)
					self.model[quantile] = qr
				# Cross-validate the quantile regression
				# TODO: this currently does not use sample_weight even if provided
				else:
					scores = []
					for alpha in self.alphas:
						qrcand = lm.QuantileRegressor(
							alpha=alpha, quantile=quantile,
						)
						score = cross_validate(
							qrcand, X, y, cv=3,
							scoring=make_scorer(d2_pinball_score, alpha=quantile),
						)
						scores.append(score['test_score'].mean())
					qr = lm.QuantileRegressor(
						alpha=self.alphas[np.argmax(scores)], quantile=quantile, 
					)
					qr.fit(features, y)
					self.model[quantile] = qr
			else:
				self.model[quantile] = None
				self.scores[quantile] = None
			
	def predict(self, X: np.array, W: np.array, Z: Optional[np.array]=None):

		features = self.feature_transform(W, X=X, Z=Z)
		all_preds = [] 
		for quantile in self.quantiles:
			if quantile == 0:
				preds = np.zeros(len(features)) + self.ymin
			elif quantile == 1:
				preds = np.zeros(len(features)) + self.ymax
			else:
				preds = self.model[quantile].predict(features)
			all_preds.append(preds)

		all_preds = np.stack(all_preds, axis=1) # n x self.nq
		# sort to ensure coherence
		all_preds = np.sort(all_preds, axis=1)
		# take centers and return
		yvals = (all_preds[:, 1:] + all_preds[:, 0:-1]) / 2
		probs = np.stack([self.probs for _ in range(len(X))], axis=0)
		return utilities.BatchedCategorical(
			vals=yvals, probs=probs
		)

class BinaryDistReg(DistReg):
	"""
	Binary regression which inherits from ``DistReg``

	Parameters
	----------
	model_type : str or sklearn class.
		Str specifying a sklearn model class to use; options include
		'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'. One can
		also directly pass an sklearn class, e.g., 
		``model_type=sklearn.linear_model.LogisticRegressionCV``.
	how_transform : str
		Str specifying how to transform the features before fitting
		the underlying model. One of several options:

		- 'identity': does not transform the features
		- 'intercept': adds an intercept
		- 'interactions' : adds treatment-covariate interactions

		The default is ``interactions``.
	montonicity : bool
		If True, ensures :math:`P(Y_i(1) = 1 | X_i) - P(Y_i(0) = 1 | X_i)` >= 0.
	monotonicity_margin : float
		When ``self.monotonicity = True``, ensures that
		:math:`P(Y_i(1) = 1 | X_i) - P(Y_i(0) = 1 | X_i)` >= margin. 
		This is important for numerical stability but does not affect 
		validity.
	model_kwargs : dict
		kwargs to sklearn model class.
	"""
	def __init__(
		self, 
		model_type: Union[str, sklearn.base.BaseEstimator] = 'logistic', 
		monotonicity: bool = False,
		monotonicity_margin: float=0.005,
		how_transform: str = 'interactions',
		**model_kwargs
	) -> None:
		self.mtype_raw = model_type
		self.model_type = parse_model_type(model_type, discrete=True)
		self.monotonicity = monotonicity
		self.margin = monotonicity_margin
		self.how_transform = str(how_transform).lower()
		self.model_kwargs = model_kwargs

		## Default kwargs
		if self.mtype_raw == 'lasso':
			self.model_kwargs['penalty'] = 'l1'
			self.model_kwargs['solver'] = self.model_kwargs.get('solver', 'saga')
		elif self.mtype_raw in ['ridge']:
			self.model_kwargs['penalty'] = 'l2'
			self.model_kwargs['solver'] = self.model_kwargs.get('solver', 'lbfgs')
		elif self.mtype_raw in ['elasticnet', 'elastic']:
			self.model_kwargs['penalty'] = 'elasticnet'
			self.model_kwargs['solver'] = self.model_kwargs.get('solver', 'saga')
			self.model_kwargs['l1_ratios'] = np.array([0, 0.5, 1])


	def fit(
		self,
		W: np.array, 
		X: np.array,
		y: np.array,
		Z: Optional[np.array]=None,
		sample_weight: Optional[np.array]=None,
	) -> None:
		self._iv = Z is not None
		# check y is binary
		if set(np.unique(y).tolist()) != set([0,1]):
			raise ValueError(f"y must be binary; instead np.unique(y)={np.unique(y)}")
		# fit 
		features = self.feature_transform(W=W, X=X, Z=Z)
		self.model = self.model_type(**self.model_kwargs)
		# Possibly use sample weights
		fit_kwargs = dict() if sample_weight is None else dict(sample_weight=sample_weight)
		self.model.fit(features, y, **fit_kwargs)

	def predict_proba(
		self, X: np.array, W: np.array, Z: Optional[np.array]=None,
	):
		# return P(Y = 1 | X, W)
		return self.model.predict_proba(
			self.feature_transform(W, X, Z=Z)
		)[:, 1]

	def predict(self, X: np.array, W: np.array, Z: Optional[np.array]=None):
		return BatchedCategorical.from_binary_probs(
			self.predict_proba(X=X, W=W, Z=Z)
		)

	def predict_counterfactuals(self, X: np.array):
		n = len(X)
		ydists = DistReg.predict_counterfactuals(self, X=X)
		## enforce monotonicity
		if not self.monotonicity:
			return ydists

		# Case 1: instrument variables (4 counterfactuals)
		if self._iv:
			for z in [0,1]:
				ydists[z] = _enforce_monotonicity(
					ydists[z][0], ydists[z][1], margin=self.margin
				)
			return ydists
		# Case 2: intent to treat (2 counterfactuals)
		else:
			return _enforce_monotonicity(ydists[0], ydists[1], margin=self.margin)


class MonotoneLogisticReg:
	"""
	A logistic regression solver which ensures that beta[0] >= 0.
	Useful for computing bounds which assume monotonicity.

	Notes
	-----
	This is meant to be used a the underlying model in a
	 :class:`BinaryDistReg` object.
	"""
	def __init__(self, lmda: float=0.001):
		self.lmda = lmda

	def fit(self, X: np.array, y: np.array, solver='CLARABEL', **solver_kwargs):
		"""
		Fits the linear model using a cvxpy backend.

		Parameters
		----------
		X : np.array
			(n,p)-shaped design matrix.
		y : np.array
			n-length vector of binary response.
		solver : str
			Solver for cvxpy to use. Default: CLARABEL.
		solver_kwargs : dict
			Kwargs for the cvxpy solver.
		"""
		# Set up data and losses
		n, p = X.shape
		sig1 = X[:, 0].std()
		zeros = np.zeros(n)
		beta = cp.Variable(p)
		X1beta = X @ beta
		term1 = cp.multiply(y, X1beta)
		term2 = cp.log_sum_exp(cp.vstack([zeros, X1beta]), axis=0)
		term3 = n * self.lmda * cp.sum(cp.power(beta[1:], 2))
		# Objective and problem
		obj = cp.Maximize(cp.sum(term1 - term2) - term3)
		problem = cp.Problem(objective=obj, constraints=[beta[0] >= 0])
		problem.solve(solver=solver, **solver_kwargs)
		# Learn value
		self.beta = beta.value

	def predict_proba(self, X: np.array):
		"""
		Probability predictions for the response.

		Parameters
		----------
		X : np.array
			(n,p)-shaped design matrix.

		Returns
		-------
		probs : np.array
			(n,2)-shaped array of probabilities where
			probs[i, k] = P(Y = k | X = X[i]) for k in {0,1}.
		"""
		mu = X @ self.beta
		p1s = np.exp(mu) / (1 + np.exp(mu))
		return np.stack([1 - p1s, p1s], axis=1)


import time
import numpy as np
import scipy as sp
from scipy import stats
from sklearn import linear_model as lm
import unittest
import pytest
import os
import sys
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import dist_reg

def convert_norm_dists_to_array(dists):
	mus = []; sigmas = []
	for dist in dists:
		mu, sigma = dist.stats()
		mus.append(mu); sigmas.append(sigma)
	return np.concatenate(mus, axis=0), np.concatenate(sigmas, axis=0)

class TestDistReg(unittest.TestCase):
	"""
	tests
	"""
	def test_oos_ridge(self):
		# create data
		np.random.seed(123)
		n, p = 100, 50
		X = np.random.randn(n, p)
		beta = np.random.randn(p) * 0.2
		y = np.random.randn(n) + X @ beta + 10
		# Fit various ridge models and oos residuals
		for fit_intercept in [True, False]:
			for alpha in [0.1, 9]:
				# Fit ridge
				kwargs = dict(fit_intercept=fit_intercept)
				ridge_nocv = lm.Ridge(alpha=alpha, solver='cholesky', **kwargs)
				ridge_nocv.fit(X, y)
				# Fit ridge CV
				ridge_cv = lm.RidgeCV(**kwargs)
				ridge_cv.fit(X, y)
				for ridge, cv in zip([ridge_nocv, ridge_cv], [False, True]):
					# Fetch out of sample
					loo_resids = dist_reg.ridge_loo_resids(
						features=X, y=y, ridge_cv_model=ridge,
					)
					# Fit ridge leaving one ind out
					ind = 3
					negind = np.array([i for i in range(n) if i != ind])
					if isinstance(ridge, lm.RidgeCV):
						ridge_negind = lm.Ridge(alpha=ridge.alpha_, **kwargs)
					else:
						ridge_negind = lm.Ridge(alpha=alpha, **kwargs)
					ridge_negind.fit(X[negind], y[negind])
					expected = y[[ind]] - ridge_negind.predict(X[[ind]])
					# test equality
					np.testing.assert_array_almost_equal(
						expected,
						loo_resids[[ind]],
						decimal=3 if not fit_intercept else 1,
						err_msg=f"Analytical ridge LOO comp failed for kwargs={kwargs}, cv={cv}"
					)


	def test_crossfit_linreg(self):
		# create model 
		np.random.seed(123)
		n = 200; p = 30; nfolds = 3; cv = 3
		X = np.random.randn(n, p)
		W = np.random.binomial(1, 0.5, size=n)
		beta = np.random.randn(p)
		Y = X @ beta + np.random.randn(n)
		# cross-fitting approach one
		model_kwargs = dict(model_type='ridge', cv=cv, how_transform='identity', eps_dist='gaussian')
		rdr = dist_reg.CtsDistReg(**model_kwargs)
		(preds0, preds1), _, _ = dist_reg.cross_fit_predictions(
			W=W, y=Y, X=X, nfolds=nfolds, model=rdr,
		)
		mu0s, sigma0s = convert_norm_dists_to_array(preds0)
		mu1s, sigma1s = convert_norm_dists_to_array(preds1)

		# manually perform cross-fitting in this case
		starts, ends = dist_reg.create_folds(n=n, nfolds=nfolds)
		p0s = []; p1s = []
		for start, end in zip(starts, ends):
			not_in_fold = [i for i in range(n) if i < start or i >= end]
			model = dist_reg.CtsDistReg(**model_kwargs)
			model.fit(W=W[not_in_fold], X=X[not_in_fold], y=Y[not_in_fold])
			pr0s = model.predict(X=X[start:end], W=np.zeros(end-start))
			pr1s = model.predict(X=X[start:end], W=np.ones(end-start))
			p0s.append(pr0s); p1s.append(pr1s)
		mu0s_exp, sigma0s_exp = convert_norm_dists_to_array(p0s)
		mu1s_exp, sigma1s_exp = convert_norm_dists_to_array(p1s)

		# check everything is the same
		for actual, expected, name in zip(
			[mu0s, sigma0s, mu1s, sigma1s],
			[mu0s_exp, sigma0s_exp, mu1s_exp, sigma1s_exp],
			['mu0', 'mu1', 'sigma0', 'sigma1']
		):
			np.testing.assert_array_almost_equal(
				actual, expected, decimal=3, err_msg=f"{name} from cross_fit_predictions did not match manual method"
			)

if __name__ == "__main__":
	# Run all tests---useful if using cprofilev
	basename = os.path.basename(os.path.abspath(__file__))
	if sys.argv[0] == f'test/{basename}':
		time0 = time.time()
		context.run_all_tests([TestDistReg()])
		elapsed = np.around(time.time() - time0, 2)
		print(f"Finished running all tests at time={elapsed}")

	# Else let unittest handle this for us
	else:
		unittest.main()



import time
import numpy as np
import pandas as pd
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
import sklearn.linear_model as lm
import sklearn.ensemble
import sklearn.neighbors

try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import lee, utilities, gen_data
from dualbounds.utilities import BatchedCategorical

def old_conditional_lee_bound(
	ps0,
	ps1,
	y0_vals,
	py0_given_s0,
	y1_vals,
	py1_given_s1,
	lower=True,
):
	"""
	Depreciated, used to test the new version.
	
	Parameters
	----------
	ps0 : np.array
		2-length array where ps0[i] = P(S(0) = i | X)
	ps1 : np.array
		2-length array where ps1[i] = P(S(1) = i | X)
	y0_vals : np.array
		nvals-length array of values y1 can take.
	py0_given_s0 : np.array
		(nvals)-length array where
		py0_given_s0[i] = P(Y(0) = yvals[i] | S(0) = 1)
	y1_vals : np.array
		nvals-length array of values y1 can take.
	py1_given_s1 : np.array
		(nvals)-length array where
		py1_given_s1[i] = P(Y(1) = yvals[i] | S(1) = 1)
	"""
	if np.any(y1_vals != np.sort(y1_vals)):
		raise ValueError(f"y1_vals must be sorted, not {y1_vals}")

	if not lower:
		return -1 * old_conditional_lee_bound(
			ps0=ps0, 
			ps1=ps1,
			y0_vals=-1 * y0_vals,
			py0_given_s0=py0_given_s0,
			y1_vals=np.flip(-1*y1_vals),
			py1_given_s1=np.flip(py1_given_s1),
			lower=True
		)

	# verify monotonicity
	if ps0[1] > ps1[1]: 
		raise ValueError(f"Monotonicity is violated, ps0={ps0}, ps1={ps1}") 

	# always takers share 
	p0 = ps0[1] / ps1[1]

	# compute E[Y(1) | Y(1) >= Q(p0), S(1)=1]
	# where Q is the quantile fn of Y(1) | S(1) = 1.
	cum_cond_probs = np.cumsum(py1_given_s1)
	cum_cond_probs /= py1_given_s1.sum() # just in case
	k = np.argmin(cum_cond_probs <= p0)
	# shave off probability from the last cell bc of discreteness
	gap = cum_cond_probs[k] / p0 - 1
	cond_probs = py1_given_s1[0:(k+1)] / p0
	cond_probs[-1] -= gap
	if np.abs(np.sum(cond_probs) - 1) > 1e-8: 
		raise ValueError(f"Cond probs sum to {np.sum(cond_probs)} != 1")
	term1 = y1_vals[0:(k+1)] @ cond_probs
	return term1 - y0_vals @ py0_given_s0

def compute_cvar_samples(dists, n, alpha, lower=True, reps=100000):
	"""
	Batched computation of 
	E[Y | Y <= Q_{alpha}(Y)] from Y ~ dists if lower = True.
	Used to test that the other cvar implementation (which is more
	efficient) is accurate.
	"""
	if isinstance(alpha, float) or isinstance(alpha, int):
		alpha = alpha * np.ones(n)
	# sample
	samples = dists.rvs(size=(reps, n))
	# Compute estimates from samples
	cvar_est = np.zeros(n)
	for i in range(n):
		samplesi = samples[:, i]
		# estimated quantile
		hatq = np.quantile(samplesi, alpha[i])
		if lower:
			cvar_est[i] = samplesi[samplesi <= hatq].mean()
		else:
			cvar_est[i] = samplesi[samplesi >= hatq].mean()
	# Return
	return cvar_est

class TestLeeHelpers(unittest.TestCase):
	"""
	tests
	"""
	def test_cvar(self):
		# create 
		np.random.seed(123)
		n = 20
		sigmas = np.random.uniform(size=n)
		mu = np.random.randn(n)
		ydists = stats.expon(loc=mu, scale=sigmas)
		# test
		for lower in [True, False]:
			avec = np.random.uniform(0.1, 0.9, size=n)
			for alpha in [0.1, 0.3, 0.5, 0.7, 0.9, avec]:
				cvar_est = lee.compute_cvar(
					ydists, n=n, alpha=alpha, lower=lower
				)
				cvar_est_samp = compute_cvar_samples(
					ydists, n=n, alpha=alpha, lower=lower
				)
				err_ratio = np.sum((cvar_est - cvar_est_samp)**2) / np.sum(cvar_est**2)
				self.assertTrue(
					err_ratio  <= 1e-3,
					f"cvar_est={cvar_est} != {cvar_est_samp}."
				)

	def test_analytical_lee_bnds(self):
		# Create dgp
		n = 40
		nvals = 20
		s0_probs = np.random.uniform(0, 0.9, size=n)
		s1_probs = s0_probs + 0.1
		# create vals
		y0_vals = np.sort(np.random.randn(n, nvals), axis=1)
		y1_vals = np.sort(np.random.randn(n, nvals), axis=1)
		y0_probs = np.random.uniform(size=(n, nvals))
		y0_probs /= y0_probs.sum(axis=1).reshape(-1, 1)
		y1_probs = np.random.uniform(size=(n, nvals))
		y1_probs /= y1_probs.sum(axis=1).reshape(-1, 1)
		# compute lee bounds---old variant
		lbounds1 = np.zeros(n)
		ubounds1 = lbounds1.copy()
		for i in range(n):
			ps0 = np.array([1 - s0_probs[i], s0_probs[i]])
			ps1 = np.array([1 - s1_probs[i], s1_probs[i]])
			args = dict(
				ps0=ps0, ps1=ps1,
				y0_vals=y0_vals[i],
				y1_vals=y1_vals[i],
				py0_given_s0=y0_probs[i],
				py1_given_s1=y1_probs[i],
			)
			lbounds1[i] = old_conditional_lee_bound(
				**args, lower=True
			)
			ubounds1[i] = old_conditional_lee_bound(
				**args, lower=False
			)
		abounds1 = np.stack([lbounds1, ubounds1], axis=0)

		# new variant
		new_args = dict(
			s0_probs=s0_probs, s1_probs=s1_probs,
			y0_probs=y0_probs, y1_probs=y1_probs,
			y1_vals=y1_vals, y0_vals=y0_vals,
		)
		_, abounds = lee.compute_analytical_lee_bound(**new_args, m=10000)

		# assert equality
		np.testing.assert_array_almost_equal(
			abounds, abounds1, 
			decimal=3,
			err_msg="discrete analytical vs. cts analytical bounds do not match"
		)

	def test_lee_delta_method_se(self):
		""" Tests that we correctly estimate the SE. """
		context._run_se_computation_test(
			dim=3,
			f=lambda x, y, z: (x - y) / z,
			arg_names=['sbetas', 'skappas', 'sgammas'],
			testname='Lee bound',
			se_function=db.lee.lee_delta_method_se,
			param_shift=1,
		)

	def test_lee_bound_no_covariates(self):
		"""
		Tests that lee_bound_no_covariates gives roughly correct SEs.
		"""
		n, n_clusters = 4000, 3000
		clusters = np.random.choice(n_clusters, n, replace=True)
		def lee_bound_output(sample_seed, B):
			data = gen_data.gen_lee_bound_data(
				n=n_clusters, p=1, lmda_dist='skewnorm', sample_seed=sample_seed, r2=0, stau=1.5, betaS_norm=0,
			)
			return lee.lee_bound_no_covariates(
				outcome=data['y'][clusters], 
				treatment=data['W'][clusters], 
				propensities=data['pis'][clusters], 
				selections=data['S'][clusters],
				clusters=clusters,
				B=B,
				verbose=False,
			)
		# Estimate standard errors
		hatses = lee_bound_output(sample_seed=1, B=100)['ses']
		# Compute oracle standard erorrs
		reps = 100
		ests = np.zeros((reps, 2))
		for r in range(reps):
			ests[r] = lee_bound_output(sample_seed=r+1, B=0)['estimates']
		ses = ests.std(axis=0)
		# Check equality
		for obj1, obj2 in zip(
			[hatses, ses], [ses, hatses]
		):
			self.assertTrue(
				np.all(obj1/obj2 <= 1.2),
				msg=f"hatses={hatses}, ses={ses} are not the same for lee_bound_no_covariates with clusters"
			)

class TestDualLeeBounds(context.DBTest):

	def test_dual_lee_lp_solver(self):
		np.random.seed(1234)
		# Create dgp
		n = 11
		nvals = 1000
		s0_probs = np.random.uniform(0, 0.9, size=n)
		s1_probs = s0_probs + 0.1
		# create dists
		y0_dists = stats.norm(
			loc=0.3*np.random.randn(n), scale=np.random.uniform(0.1, 1, size=n),
		)
		y1_dists = stats.expon(
			loc=0.3*np.random.randn(n), scale=np.random.uniform(0.1, 1, size=n),
		)
		y = y1_dists.rvs() # placeholder, not important
		# args
		args = dict(
			s0_probs=s0_probs, s1_probs=s1_probs, y1_dists=y1_dists,
		)
		# analytical solution
		_, expected = lee.compute_analytical_lee_bound(**args, y0_dists=y0_dists)
		# analytical bounds based on linear programming
		ldb = lee.LeeDualBounds(
			outcome=y, treatment=np.ones(n), selections=np.random.binomial(1, 0.5, n),
		)
		ldb.compute_dual_variables(
			**args, ymin=-10, ymax=10, nvals=nvals
		)
		lp_bounds = ldb.objvals - ldb.dxs - y0_dists.mean() * s0_probs
		lp_bounds = lp_bounds / s0_probs
		np.testing.assert_array_almost_equal(
			lp_bounds,
			expected,
			decimal=2,
			err_msg="Lee LP bounds do not agree with analytical bounds",
		)
		# Test that dx == 0
		np.testing.assert_array_almost_equal(
			ldb.dxs,
			np.zeros(ldb.dxs.shape),
			decimal=5,
			err_msg=f"Lee bound dxs are too large"
		)

	def test_dual_lee_ipw_bounds(self):
		np.random.seed(1234)
		# Create dgp
		n = 50
		nvals = 100
		s0_probs = np.random.uniform(0, 0.9, size=n)
		s1_probs = s0_probs + 0.1
		# create dists
		for eps_dist in ['bernoulli', 'expon']:
			y0_dists = utilities.parse_dist(
				eps_dist, loc=np.random.randn(n),
				scale=np.random.uniform(0.1, 1, size=n),
			)
			y1_dists = utilities.parse_dist(
				eps_dist, loc=np.random.randn(n), 
				scale=np.random.uniform(0.1, 1, size=n),
			)
			if eps_dist == 'bernoulli':
				y1_dists_input = BatchedCategorical.from_scipy(y1_dists)
			else:
				y1_dists_input = y1_dists

			args = dict(
				s0_probs=s0_probs, s1_probs=s1_probs, y1_dists=y1_dists_input,
			)
			# sample data. a trick to make the test run faster
			# is to sample many values of y per dual variable
			N = 2000 # num samples per value of x
			pis = np.ones(n) / 2
			W = np.random.binomial(1, pis, size=(N,n))
			Y0 = y0_dists.rvs(size=(N,n))
			Y1 = y1_dists.rvs(size=(N,n))
			Y = Y0.copy(); Y[W == 1] = Y1[W == 1]
			S0 = np.random.binomial(1, s0_probs, size=(N,n))
			S1 = np.random.binomial(1, s1_probs, size=(N,n))
			S = S0.copy(); S[W == 1] = S1[W == 1]

			# analytical solution
			_, abounds = lee.compute_analytical_lee_bound(**args, y0_dists=y0_dists)
			# convert to E[Y(1) S(0)]
			expected = np.mean(abounds * s0_probs + s0_probs * y0_dists.mean(), axis=1)
			# compute dual variables
			ldb = lee.LeeDualBounds(
				outcome=Y[0], selections=S[0], treatment=W[0], X=None, propensities=pis
			)
			ldb.y0_dists = y0_dists
			ldb.s0_probs = s0_probs
			ldb.compute_dual_variables(**args, ymin=-5, ymax=5)
			# compute IPW/AIPW summands
			ipws = []
			aipws = []
			for i in range(N):
				# use same dual variables, new data
				ldb.y = Y[i]
				ldb.S = S[i]
				ldb.W = W[i]
				ldb._compute_realized_dual_variables(y=Y[i], S=S[i])
				ldb._compute_ipw_summands()
				ipws.append(ldb.ipw_summands)
				aipws.append(ldb.aipw_summands)
			objval = ldb.objvals.mean(axis=1)
			ipw_ests = np.concatenate(ipws, axis=1).mean(axis=1)
			aipw_ests = np.concatenate(aipws, axis=1).mean(axis=1)
			for method, ests in zip(['ipw', 'aipw', 'LP'], [ipw_ests, aipw_ests, objval]):
				np.testing.assert_array_almost_equal(
					ests,
					expected,
					decimal=2,
					err_msg=f"{method} bounds do not agree with analytical bounds",
				)

	def test_from_pd(self):
		# Sample data
		data = db.gen_data.gen_lee_bound_data(n=50, p=5, eps_dist='bernoulli', r2=0)

		# Method 1
		ldb1 = db.lee.LeeDualBounds(
			covariates=pd.DataFrame(data['X']),
			outcome=pd.Series(data['y']),
			treatment=pd.Series(data['W']),
			selections=pd.Series(data['S']),
			propensities=pd.DataFrame(data['pis']),
		)

		# Method 2
		ldb2 = db.lee.LeeDualBounds(
			covariates=data['X'],
			outcome=data['y'],
			treatment=data['W'],
			selections=data['S'],
			propensities=data['pis'],
		)

		# Test equality
		self.assertTrue(
			ldb1.X.shape == ldb2.X.shape,
			"LeeDualBounds init. from pandas changes the shape of the covariates"
		)
		for expected, out, name in zip(
			[ldb2.y, ldb2.W, ldb2.pis, ldb2.S],
			[ldb1.y, ldb1.W, ldb1.pis, ldb1.S],
			['y', 'W', 'pis', 'S'],
		):
			np.testing.assert_array_almost_equal(
				expected,
				out,
				decimal=8,
				err_msg=f"LeeDualBounds init. from pandas changes {name} values."
			)

	def test_lee_bound_clustered_ses(self):
		data = db.gen_data.gen_lee_bound_data(
			n=300, p=1, r2=0, dgp_seed=1, sample_seed=1, lmda_dist='powerlaw',
			stau=1, betaS_norm=0,
		)
		# Stack data
		data = np.stack(
			[data['y'], data['W'], data['S'], data['pis'], data['X'][:, 0]],
			axis=1
		)
		def lee_se_function(data, clusters):
			ldb = lee.LeeDualBounds(
				outcome=data[:, 0],
				treatment=data[:, 1],
				selections=data[:, 2],
				propensities=data[:, 3],
				covariates=data[:, 4].reshape(-1, 1),
				clusters=clusters,
				selection_model='ridge',
			)
			ldb.fit().summary()
			return ldb.ses
		self.check_clustered_ses(func=lee_se_function, data=data, msg_context='Lee bounds')

	def test_model_type_inputs(self):
		"""
		Tests that the various model classes don't error and give
		the correct underlying model type.
		"""
		data = db.gen_data.gen_lee_bound_data(
			n=500, p=3, r2=0, dgp_seed=1, sample_seed=1,
		)

		for outcome_model, selection_model, propensity_model, pis in zip(
			['rf', 'ridge', 'knn', 'elastic'],
			['monotone_logistic', 'knn', 'logistic', 'rf'],
			[None, None, 'knn', 'logistic'],
			[data['pis'], data['pis'], None, None]
		):
			ldb = lee.LeeDualBounds(
				outcome=data['y'], treatment=data['W'], selections=data['S'], covariates=data['X'], pis=pis,
				outcome_model=outcome_model, selection_model=selection_model, propensity_model=propensity_model,
			)
			ldb.fit(nfolds=3).summary()
			# Test y-model is correct
			Ym = ldb.model_fits[0].model
			Y_exp = db.dist_reg.parse_model_type(outcome_model, discrete=False)
			self.assertTrue(
				isinstance(Ym, Y_exp),
				f"fit model {Ym} with outcome_model={outcome_model} is not of type {Y_exp}"
			)
			# Test s-model is correct
			Sm = ldb.selection_model_fits[0].model
			S_exp = db.dist_reg.parse_model_type(selection_model, discrete=True)
			self.assertTrue(
				isinstance(Sm, S_exp),
				f"fit model {Sm} with selection_model={selection_model} is not of type {S_exp}"
			)
			# Test pi model is correct if fit
			if pis is None:
				Wm = ldb.propensity_model_fits[0]
				W_exp = db.dist_reg.parse_model_type(propensity_model, discrete=True)
				self.assertTrue(
					isinstance(Wm, W_exp),
					f"fit model {Wm} with selection_model={propensity_model} is not of type {W_exp}"
				)


	@pytest.mark.slow
	def test_lee_consistency(self):
		n = 10000
		p = 3
		r2 = 0.0
		tau = 2
		for eps_dist in ['bernoulli', 'gaussian']:
			discrete = eps_dist == 'bernoulli'
			data = db.gen_data.gen_lee_bound_data(
				n=n, p=p, r2=r2, tau=tau, dgp_seed=1, sample_seed=1,
				eps_dist=eps_dist,
			)
			y0_dists = data['y0_dists']
			y1_dists = data['y1_dists']
			oracle_args = dict(
				s0_probs=data['s0_probs'], 
				s1_probs=data['s1_probs'],
				y0_dists=BatchedCategorical.from_scipy(y0_dists) if discrete else y0_dists, 
				y1_dists=BatchedCategorical.from_scipy(y1_dists) if discrete else y1_dists,
			)
			## Ground truth
			expected, _ = lee.compute_analytical_lee_bound(**oracle_args)
			## Oracle test
			ldb_oracle = lee.LeeDualBounds(
				outcome=data['y'], treatment=data['W'], selections=data['S'], covariates=data['X'], propensities=data['pis'],
			)
			est_oracle = ldb_oracle.fit(**oracle_args, suppress_warning=True).estimates
			## Actual dual bounds
			ldb = lee.LeeDualBounds(
				outcome=data['y'], treatment=data['W'], selections=data['S'], covariates=data['X'], propensities=data['pis'],
			)
			est_actual = ldb.fit(nfolds=3).estimates
			for est, name in zip([est_oracle, est_actual], ['Oracle', 'Dual']):
				np.testing.assert_array_almost_equal(
					est,
					expected,
					decimal=1,
					err_msg=f"{name} Lee bound is not consistent with n={n}, eps_dist={eps_dist}"
				)


if __name__ == "__main__":
	# Run all tests---useful if using cprofilev
	basename = os.path.basename(os.path.abspath(__file__))
	if sys.argv[0] == f'test/{basename}':
		time0 = time.time()
		context.run_all_tests([TestHelpers(), TestDualLeeBounds()])
		elapsed = np.around(time.time() - time0, 2)
		print(f"Finished running all tests at time={elapsed}")

	# Else let unittest handle this for us
	else:
		unittest.main()



import time
import numpy as np
import pandas as pd
import scipy as sp
from scipy import stats
import unittest
import pytest
import os
import sys
import sklearn.ensemble
from sklearn import linear_model as lm
try:
	from . import context
	from .context import dualbounds as db
# For profiling
except ImportError:
	import context
	from context import dualbounds as db

from dualbounds import generic, utilities, gen_data
from dualbounds.utilities import parse_dist, BatchedCategorical


class TestGenericDualBounds(context.DBTest):

	def test_var_ite_oracle(self):
		"""
		Tests that Var(ITE) objvals are accurate.
		"""
		# create dists
		np.random.seed(123)
		n = 1
		for dist0, dist1, discrete, support in zip(
			['gaussian', 'expon', 'tdist', 'bernoulli'],
			['gaussian', 'gaussian', 'expon', 'bernoulli'],
			[False, False, False, True],
			[None, None, None, np.array([0,1])],
		):
			y0_dists = parse_dist(
				dist0, loc=np.random.randn(n), scale=np.random.uniform(0.1, 1, size=n)
			)
			y1_dists = parse_dist(
				dist1, loc=np.random.randn(n), scale=np.random.uniform(0.1, 1, size=n)
			)
			y0_dists_input = BatchedCategorical.from_scipy(y0_dists) if discrete else y0_dists
			y1_dists_input = BatchedCategorical.from_scipy(y1_dists) if discrete else y1_dists
			# analytically compute lower/upper bounds on var(ITE)
			reps = 100000
			U = np.random.uniform(size=reps)
			y1 = y1_dists.ppf(U); y0l = y0_dists.ppf(U); y0u = y0_dists.ppf(1-U)
			lower = np.std(y1 - y0l)**2
			upper = np.std(y1 - y0u)**2
			mu2 = (y1_dists.mean() - y0_dists.mean())**2
			# create dualbound.
			# the input data is a dummy variable that does nothing
			nvals = 2000
			vdb = db.generic.DualBounds(
				f=lambda y0, y1, x: (y0-y1)**2,
				covariates=np.zeros(1),
				treatment=np.zeros(1),
				propensities=np.ones(1) * 1/2,
				outcome=np.zeros(1),
				discrete=discrete,
				support=support,
			)
			vdb.compute_dual_variables(
				y0_dists=y0_dists_input, 
				y1_dists=y1_dists_input, 
				y0_min=np.min(y0l),
				y0_max=np.max(y0l),
				y1_min=np.min(y1), 
				y1_max=np.max(y1),
				nvals0=nvals,
				nvals1=nvals,
			)
			# check if objvals match with lower, upper
			np.testing.assert_array_almost_equal(
				np.array([lower, upper]),
				vdb.objvals[:, 0] - mu2,
				decimal=1.5,
				err_msg=f"LP bounds for {dist0, dist1} do not agree with analytical bounds on Var(ITE)",
			)

	def test_gridsearch_loss(self):
		# Create fake heavy-tailed data
		np.random.seed(123)
		n = 300; nvals = 50; grid_size = 3 * nvals
		y0_dists = parse_dist(
			'tdist', loc=np.random.randn(n), scale=np.random.uniform(0.1, 1, size=n)
		)
		y1_dists = parse_dist(
			'expon', loc=np.random.randn(n), scale=np.random.uniform(0.1, 1, size=n)
		)
		y0 = y0_dists.rvs()
		y1 = y1_dists.rvs()
		pis = np.ones(n) / 2
		W = np.random.binomial(1, pis, size=n)
		Y = y0.copy(); Y[W == 1] = y1[W == 1].copy()
		X = np.random.randn(n, 1)
		# Try with many different fs
		fs = [
			lambda y0, y1, x: (y0-y1)**2,
			lambda y0, y1, x: y0 <= y1,
			lambda y0, y1, x: (y0 <= 0) * (y1 <= 0),
		]
		# Test the size of objdiffs is small on average
		for f in fs:
			gdb = db.generic.DualBounds(
				f=f,
				covariates=X,
				treatment=W,
				propensities=pis,
				outcome=Y,
			)
			gdb.compute_dual_variables(
				y0_dists=y0_dists, 
				y1_dists=y1_dists,
				nvals0=nvals,
				nvals1=nvals,
			)
			np.testing.assert_array_almost_equal(
				gdb.objdiffs.mean(axis=1),
				np.zeros(2),
				decimal=1.5,
				err_msg=f"For f={f}, objdiffs are too large."
			)


	def test_optimal_dualvars_consistency(self):
		"""
		Tests that if we evaluate the optimal dual variables
		we get the same value as the objvals in the perfectly
		well-specified case.
		"""
		np.random.seed(1)
		n, nvals, reps = 11, 5, 2000
		# propensity scores
		pis = np.random.uniform(0.3, 0.7, size=n)
		# Create distributions satisfying monotonicity
		y0_vals = np.random.randn(n, nvals)
		#y0_vals = np.stack([y0_vals for _ in range(n)], axis=0)
		probs = np.random.uniform(size=(n, nvals))
		probs /= probs.sum(axis=-1).reshape(n, 1)
		y1_vals = y0_vals + 1
		# create distributions
		y0_dists = BatchedCategorical(vals=y0_vals, probs=probs)
		y1_dists = BatchedCategorical(vals=y1_vals, probs=probs)
		ydists_input = dict(
			y0_dists=y0_dists, y1_dists=y1_dists, suppress_warning=True,
		)
		# Sample values of Y and W
		W = np.random.binomial(1, pis, size=(reps, n)).astype(int)
		Y = np.zeros((reps, n))
		for r in range(reps):
			y = y0_dists.rvs()
			y[W[r] == 1] = y1_dists.rvs()[W[r] == 1]
			Y[r] = y
		# Define estimand
		f = lambda y0, y1, x : y1 - y0 >= 0.5
		# Loop through different support restrictions
		for srname, support_restriction in zip(
			['none', 'monotonicity'], 
			[None, lambda y0, y1, x: y0 <= y1]
		):
			for dual_strategy in ['ot', 'lp', 'qp', 'se']:
				# Fit original dual bounds object
				gdb = db.generic.DualBounds(
					f=f,
					outcome=Y[0],
					treatment=W[0],
					propensities=pis,
					discrete=True,
					support_restriction=support_restriction,
				)
				gdb.fit(
					dual_strategy=dual_strategy, **ydists_input
				)#.summary()
				# Now, evaluate many dual variables
				aipw_estimates = np.zeros((reps, 2))
				ipw_estimates = np.zeros((reps, 2))
				for r in range(reps):
					# Collect W and Y
					# Compute new (A)IPW summands 
					gdb.W = W[r]
					gdb._compute_realized_dual_variables(y=Y[r], grid_size=0)
					gdb._compute_final_bounds(aipw=True)
					aipw_estimates[r] = gdb.estimates
					gdb._compute_final_bounds(aipw=False)
					ipw_estimates[r] = gdb.estimates
				# Test
				for method, estimates in zip(
					['aipw', 'ipw '], [aipw_estimates, ipw_estimates]
				):
					np.testing.assert_array_almost_equal(
						estimates.mean(axis=0),
						gdb.objvals.mean(axis=1),
						decimal=2 if method == 'aipw' else 1,
						err_msg=f"Oracle {method} estimates are not unbiased for objective value, dual_strategy={dual_strategy}"
					)

	def test_base_models_cts(self):
		"""
		Ensure the built-in model types work.
		"""
		# Generate data
		data = gen_data.gen_regression_data(n=50, p=5, heterosked='exp_linear')
		f = lambda y0, y1, x: y0 <= y1
		## 1. Model types to test
		## 1a. Stringkwargs
		model_types = ['ridge', 'lasso', 'elasticnet', 'rf', 'knn']
		# expected types
		mt_exps = [
			db.dist_reg.parse_model_type(mt, discrete=False) for mt in model_types
		]
		## 1b. Other inputs
		model_types.append(db.dist_reg.CtsDistReg(model_type='ridge'))
		model_types.append(db.dist_reg.QuantileDistReg(nquantiles=10, alphas=[0, 0.1]))
		mt_exps.extend([lm.RidgeCV, dict])

		## 2. sigma_models to test
		sigma_model_types = ['none', 'lasso', 'ols', 'None', None, None, None]
		sigma_model_exps = []
		for m2t in sigma_model_types:
			if m2t is None:
				sigma_model_exps.append(None)
			elif m2t.lower() == 'none':
				sigma_model_exps.append(None)
			else:
				sigma_model_exps.append(db.dist_reg.parse_model_type(m2t, discrete=False))


		## 3. Epsilon dists to test
		eps_dists = ['empirical', 'laplace', 'empirical', 'expon', 'tdist']
		eps_dists.extend(['these_args', 'should_be_ignored_and_not_raise_error'])
		for model, mt_exp, sigma_model, sigma_model_exp, eps_dist in zip(
			model_types,
			mt_exps,
			sigma_model_types,
			sigma_model_exps,
			eps_dists
		):
			gdb = db.generic.DualBounds(
				f=f, 
				covariates=data['X'], 
				treatment=data['W'], 
				outcome=data['y'],
				propensities=data['pis'], 
				outcome_model=model,
				heterosked_model=sigma_model,
				eps_dist=eps_dist,
			)
			gdb.fit(nfolds=3, alpha=0.05).summary()
			# Check the correct class for the main model
			Ym = gdb.model_fits[0].model
			self.assertTrue(
				isinstance(Ym, mt_exp),
				f"fit outcome_model {Ym} with propensity_model={model} is not of type {mt_exp}"
			)
			# Check correct class for the heteroskedastic model
			if sigma_model_exp is not None:
				Ym2 = gdb.model_fits[0].sigma2_model
				self.assertTrue(
					isinstance(Ym2, sigma_model_exp),
					f"sigma_model {Ym2} with heterosked_model={sigma_model} is not of type {sigma_model_exp}"
				)
			if sigma_model_exp is None:
				# In this case heterosked_model='none', so there should be no sigma_model
				with np.testing.assert_raises(AttributeError):
					x = gdb.model_fits[0].sigma2_model

	def test_base_models_discrete(self):
		"""
		Same as above but for discrete models.
		"""
		data = gen_data.gen_regression_data(n=300, p=3, eps_dist='bernoulli', r2=0)
		f = lambda y0, y1, x: y0 <= y1
		model_types = ['ridge', 'lasso', 'elasticnet', 'rf', 'knn']
		expected_mts = [
			db.dist_reg.parse_model_type(mt, discrete=True) for mt in model_types
		]
		model_types.append(db.dist_reg.BinaryDistReg(model_type='ridge'))
		expected_mts.append(lm.LogisticRegressionCV)
		for model_type, mt_exp in zip(model_types, expected_mts):
			gdb = db.generic.DualBounds(
				f=f, 
				covariates=data['X'], 
				treatment=data['W'], 
				outcome=data['y'],
				propensities=data['pis'], 
				outcome_model=model_type,
			)
			gdb.fit(nfolds=2, alpha=0.05).summary()
			# Check the correct model type
			Ym = gdb.model_fits[0].model
			self.assertTrue(
				isinstance(Ym, mt_exp),
				f"fit model {mt_exp} with outcome_model={model_type} is not of type {mt_exp}"
			)


	def test_fit_propensity_scores(self):
		"""
		Same as above but for propensity_model
		"""
		data = gen_data.gen_regression_data(n=300, p=3, eps_dist='gaussian', r2=0, heterosked='invnorm')
		f = lambda y0, y1, x : y0 <= y1
		propensity_models = ['ridge', 'lasso', 'knn', 'rf']
		expecteds = [
			db.dist_reg.parse_model_type(wm, discrete=True) for wm in propensity_models
		]
		propensity_models.append(sklearn.ensemble.AdaBoostClassifier(algorithm='SAMME'))
		expecteds.append(sklearn.ensemble.AdaBoostClassifier)
		for propensity_model, expected in zip(propensity_models, expecteds):
			gdb = db.generic.DualBounds(
				f=f,
				covariates=data['X'], treatment=data['W'], outcome=data['y'],
				propensities=None, # to be estimated 
				outcome_model='ridge',
				propensity_model=propensity_model,
			)
			gdb.fit(nfolds=3).summary()
			# Check the correct class
			Wm = gdb.propensity_model_fits[0]
			self.assertTrue(
				isinstance(Wm, expected),
				f"fit propensity_model {Wm} with propensity_model={propensity_model} is not of type {expected}"
			)

	def test_from_pd(self):
		# Sample data. The different heterosked kwargs are just to increase coverage
		for eps_dist, heterosked in zip(
			['gaussian', 'bernoulli'], ['linear', 'norm']
		):
			data = gen_data.gen_regression_data(
				n=100, p=5, eps_dist=eps_dist, r2=0, interactions=False, heterosked=heterosked
			)

			# Method 1
			f = lambda y0, y1, x: y1 - y0
			gdb1 = db.generic.DualBounds(
				f=f,
				outcome=pd.Series(data['y']),
				treatment=pd.Series(data['W']),
				covariates=pd.DataFrame(data['X']),
				propensities=pd.Series(data['pis']),
			).fit()

			# Method 2
			gdb2 = db.generic.DualBounds(
				f=f,
				covariates=data['X'],
				outcome=data['y'],
				treatment=data['W'],
				propensities=data['pis'],
			)

			# Test equality
			self.assertTrue(
				gdb1.X.shape == gdb2.X.shape,
				"using pandas changes the shape of the covariates"
			)
			for expected, out, name in zip(
				[gdb2.y, gdb2.W, gdb2.pis],
				[gdb1.y, gdb1.W, gdb1.pis],
				['y', 'W', 'pis'],
			):
				np.testing.assert_array_almost_equal(
					expected,
					out,
					decimal=8,
					err_msg=f"Using pandas initialization changes {name} values."
				)

		# Test handling of strings
		W = pd.Series(
			np.array(['c', 't'])[data['W'].astype(int)]
		).astype(str)
		# Create some nans to test nan handling
		data['X'][:, 1:][data['X'][:, 1:] < -2] = np.nan
		# Create some discrete features
		X = pd.DataFrame(data['X'])
		X[0] = X[0].apply(lambda x: 'a' if x > 0 else 'b')
		gdb = db.generic.DualBounds(
			f=f,
			outcome=data['y'],
			treatment=W,
			covariates=X,
			propensities=data['pis'],
		).fit(nfolds=2).summary()

	def test_generic_model_selection_no_error(self):
		# Make sure that using multiple outcome models doesn't lead to an error
		data = db.gen_data.gen_regression_data(n=100, p=1, sample_seed=123)
		gdb = db.generic.DualBounds(
			f=lambda y0, y1, x: np.maximum(y1-y0,0),
			outcome=data['y'],
			covariates=data['X'],
			treatment=data['W'],
			propensities=data['pis'],
			# outcome models with and without interactions
			outcome_model=['ridge', db.dist_reg.CtsDistReg('ridge', how_transform='identity')],
		).fit(nfolds=3)

	@pytest.mark.slow
	def test_support_restriction_consistency(self):
		tau = 1
		n = 11
		data = db.gen_data.gen_regression_data(
			n=n, p=1, r2=0.5, dgp_seed=1, sample_seed=1, eps_dist='gaussian',
			interactions=False, tau=tau,
		)
		# Train on augmented data to get consistent estimates
		def gen_new_data(reps):		
			W = np.random.binomial(1, data['pis'], size=(reps, n))
			Y0 = data['y0_dists'].rvs(size=(reps, n))
			Y1 = data['y1_dists'].rvs(size=(reps, n))
			Y = Y0.copy(); Y[W==1] = Y1[W==1].copy()
			return Y, W

		# Fit ridge
		t0 = time.time()
		reps = 5000
		Y, W = gen_new_data(reps=reps)
		ridge = db.dist_reg.CtsDistReg(model_type='ridge', how_transform='identity')
		ridge.fit(
			y=np.concatenate([Y[r] for r in range(reps)], axis=0),
			W=np.concatenate([W[r] for r in range(reps)], axis=0),
			X=np.concatenate([data['X'] for r in range(reps)], axis=0),
		)
		hat_y0_dists, hat_y1_dists = ridge.predict_counterfactuals(data['X'])


		# Fit oracle and empirical dual variables
		gdb_args = dict(
			f=lambda y0, y1, x: (y1-y0)**2,
			outcome=data['y'],
			treatment=data['W'],
			propensities=data['pis'],
			covariates=data['X'],
			support_restriction=lambda y0, y1, x: y0 <= y1,
		)
		oracle_gdb = db.generic.DualBounds(**gdb_args)
		oracle_gdb.fit(
			y0_dists=data['y0_dists'], y1_dists=data['y1_dists'], suppress_warning=True,
		).summary()
		empirical_gdb = db.generic.DualBounds(**gdb_args)
		empirical_gdb.fit(
			y0_dists=hat_y0_dists, y1_dists=hat_y1_dists, suppress_warning=True,
		).summary()

		# Evaluate many times
		reps = 500
		Y, W = gen_new_data(reps=reps)
		oracle_ests = np.zeros((reps, 2))
		empirical_ests = np.zeros((reps, 2))
		for r in range(reps):
			for ests, gdb in zip([oracle_ests, empirical_ests], [oracle_gdb, empirical_gdb]):
				gdb.W = W[r]
				gdb._compute_realized_dual_variables(y=Y[r], grid_size=0)
				gdb._compute_final_bounds()
				ests[r] = gdb.estimates
		# Check validity
		o_ests = oracle_ests.mean(axis=0)
		e_ests = empirical_ests.mean(axis=0)
		o_objs = oracle_gdb.objvals.mean(axis=1)
		e_objs = empirical_gdb.objvals.mean(axis=1)
		err_msg = f"DualBounds w/ monotonicity is not consistent: empirical objvals={e_objs}"
		err_msg += f", oracle objvals={o_objs}, empirical est={e_ests}, oracle est={o_ests}"
		for obj1, obj2 in zip(
			[e_objs, e_ests, e_ests],
			[o_objs, o_ests, o_objs],
		):
			np.testing.assert_allclose(
				obj1, 
				obj2,
				rtol=1e-1,
				err_msg=err_msg
			)

	def test_generic_dualbound_clustered_ses(self):
		data = db.gen_data.gen_regression_data(
			n=100, p=1, r2=0, dgp_seed=1, sample_seed=1, eps_dist='gaussian',
		)
		f = lambda y0, y1, x: np.abs(y1 - y0)
		# Stack data
		data = np.stack(
			[data['y'], data['W'], data['pis'], data['X'][:, 0]],
			axis=1
		)
		def gdb_se_function(data, clusters):
			gdb = generic.DualBounds(
				f=f,
				outcome=data[:, 0],
				treatment=data[:, 1],
				propensities=data[:, 2],
				covariates=data[:, 3].reshape(-1, 1),
				clusters=clusters,
			)
			gdb.fit().summary()
			return gdb.ses
		self.check_clustered_ses(func=gdb_se_function, data=data, msg_context='DualBounds')

class TestPluginBounds(unittest.TestCase):

	def test_plug_in_no_covariates(self):
		np.random.seed(123)
		## DGP
		n, reps, max_nvals = 350, 100, 150
		mu0 = np.zeros(n)
		mu1 = np.random.randn(n)
		y0_dists = stats.norm(loc=mu0)
		y1_dists = stats.norm(loc=mu1)
		pis = np.clip(np.exp(2*mu1) / (1 + np.exp(2*mu1)), 0.25, 0.75)
		## Unobserved Y(1), Y(0)
		y1 = y1_dists.rvs()
		y0 = y0_dists.rvs()
		## Target
		f = lambda y0, y1, x: np.maximum(0, y1-y0)
		## Initialize outputs
		naive = np.zeros((reps, 2))
		ests = np.zeros((reps, 2))
		oracles = np.zeros((reps, 2))
		for r in range(reps):
			# Observed Y(1), Y()
			W = np.random.binomial(1, pis, size=n)
			y = y1.copy(); y[W == 0] = y0[W==0]
			# Naive estimator
			naive[r] = db.generic.plug_in_no_covariates(
				outcome=y, 
				treatment=W, 
				f=f, 
				propensities=None, 
				max_nvals=max_nvals,
				B=0
			)['estimates']
			# Oracle estimator
			oracles[r] = db.generic.plug_in_no_covariates(
				outcome=np.concatenate([y1, y0]), 
				treatment=np.concatenate([np.ones(n), np.zeros(n)]), 
				f=f,
				propensities=None,
				max_nvals=2*max_nvals,
				B=0,
			)['estimates']
			# Real estimator
			ests[r] = db.generic.plug_in_no_covariates(
				outcome=y, 
				treatment=W,
				f=f,
				propensities=pis,
				max_nvals=max_nvals,
				B=0
			)['estimates']

		# Take averages
		naive_mu = naive.mean(axis=0)
		est_mu = ests.mean(axis=0)
		oracle_mu = oracles.mean(axis=0)

		# Test inaccuracy of naive method
		msg = "Nonconstant prop scores should invalidate plug_in_no_covariates"
		for lower in [1, 0]:
			self.assertTrue(
				np.abs(naive_mu[1-lower] - oracle_mu[1-lower]) > 0.1,
				msg + f": naive_est={naive_mu}, correct={oracle_mu}" 
			)

		# Test accuracy of correct method
		np.testing.assert_array_almost_equal(
			3*est_mu,
			3*oracle_mu,
			decimal=1,
			err_msg="plug_in_no_covariates fails to adjust for prop scores"
		)

	def test_plug_in_ses(self):
		np.random.seed(123)
		f = lambda y0, y1, x: np.abs(y0 - y1)
		sample_kwargs = dict(n=200, p=1, dgp_seed=123)
		# Compute SEs on one run
		data = gen_data.gen_regression_data(**sample_kwargs, sample_seed=1)
		output = generic.plug_in_no_covariates(
			f=f,
			outcome=data['y'],
			treatment=data['W'],
			propensities=data['pis'],
			clusters=None,
			B=200,
		)
		# Compute oracle SE on many runs
		reps = 200
		ests = np.zeros((reps, 2))
		for r in range(reps):
			data = gen_data.gen_regression_data(**sample_kwargs, sample_seed=r+1)
			ests[r] = generic.plug_in_no_covariates(
				f=f,
				outcome=data['y'],
				treatment=data['W'],
				propensities=data['pis'],
				clusters=None,
				B=0,
			)['estimates']
		# Compare
		ses = ests.std(axis=0),
		hatses = output['ses']
		for obj1, obj2 in zip([ses, hatses], [hatses, ses]):
			self.assertTrue(
				np.all(obj1 / obj2 <= 1.2),
				msg=f"plug_in_no_covariates estimated se={hatses}, truth={ses}; these are not the same"
			)





if __name__ == "__main__":
	# Run all tests---useful if using cprofilev
	basename = os.path.basename(os.path.abspath(__file__))
	if sys.argv[0] == f'test/{basename}':
		time0 = time.time()
		context.run_all_tests([TestPluginBounds(), TestGenericDualBounds()])
		elapsed = np.around(time.time() - time0, 2)
		print(f"Finished running all tests at time={elapsed}")

	# Else let unittest handle this for us
	else:
		unittest.main()
